"""
This module contains the C++ shared object files which are accessed via python.
"""

from pycoalescence.necsim.necsim import CPSpatialSimulation, CPNSESimulation, CSpatialSimulation, CNSESimulation
